@@ -8,7 +8,7 @@ package Future::Utils;
 use strict;
 use warnings;
 
-our $VERSION = '0.22';
+our $VERSION = '0.24';
 
 use Exporter 'import';
 
@@ -34,6 +34,17 @@ C<Future::Utils> - utility functions for working with C<Future> objects
 
 =head1 SYNOPSIS
 
+ use Future::Utils qw( call_with_escape );
+
+ my $result_f = call_with_escape {
+    my $escape_f = shift;
+    my $f = ...
+       $escape_f->done( "immediate result" );
+       ...
+ };
+
+Z<>
+
  use Future::Utils qw( repeat try_repeat try_repeat_until_success );
 
  my $eventual_f = repeat {
@@ -69,6 +80,7 @@ C<Future::Utils> - utility functions for working with C<Future> objects
  } foreach => \@items;
 
 Z<>
+
  use Future::Utils qw( fmap_concat fmap_scalar fmap_void );
 
  my $result_f = fmap_concat {
@@ -233,11 +245,18 @@ The generator is called in list context but should return only one item per
 call. Subsequent values will be ignored. When it has no more items to return
 it should return an empty list.
 
+For backward compatibility this function will allow a C<while> or C<until>
+condition that requests a failure be repeated, but it will print a one-time
+warning if it has to do that. To apply repeating behaviour that can catch and
+retry failures, use C<try_repeat> instead.
+
 =cut
 
+my $try_repeat_warned;
+
 sub _repeat
 {
-   my ( $code, $return, $trialp, $cond, $sense ) = @_;
+   my ( $code, $return, $trialp, $cond, $sense, $is_try ) = @_;
 
    my $prev = $$trialp;
 
@@ -250,7 +269,7 @@ sub _repeat
          $return ||= $trial->new;
          $trial->on_ready( sub {
             return if $$trialp->is_cancelled;
-            _repeat( $code, $return, $trialp, $cond, $sense );
+            _repeat( $code, $return, $trialp, $cond, $sense, $is_try );
          });
          return $return;
       }
@@ -263,6 +282,11 @@ sub _repeat
          return $return;
       }
 
+      if( !$is_try and $trial->failure ) {
+         $try_repeat_warned or $try_repeat_warned++,
+            carp "Using Future::Utils::repeat to retry a failure is deprecated; use try_repeat instead";
+      }
+
       # redo
       undef $$trialp;
    }
@@ -332,8 +356,8 @@ sub repeat(&@)
    my $future = $args{return};
 
    my $trial;
-   $args{while} and $future = _repeat( $code, $future, \$trial, $args{while}, 0 );
-   $args{until} and $future = _repeat( $code, $future, \$trial, $args{until}, 1 );
+   $args{while} and $future = _repeat( $code, $future, \$trial, $args{while}, 0, $args{try} );
+   $args{until} and $future = _repeat( $code, $future, \$trial, $args{until}, 1, $args{try} );
 
    $future->on_cancel( sub { $trial->cancel } );
 
@@ -342,16 +366,22 @@ sub repeat(&@)
 
 =head2 $future = try_repeat { CODE } ...
 
-Currently a simple alias to C<repeat>. However, in some later version the
-C<repeat> function will be changed so that if a trial future fails, then the
-eventual future will immediately fail as well, making its semantics a little
-closer to that of a C<while {}> loop in Perl. Code that specifically wishes
-to catch failures in trial futures and retry the block should use
-C<try_repeat> specifically.
+A variant of C<repeat> that doesn't warn when the trial fails and the
+condition code asks for it to be repeated.
+
+In some later version the C<repeat> function will be changed so that if a
+trial future fails, then the eventual future will immediately fail as well,
+making its semantics a little closer to that of a C<while {}> loop in Perl.
+Code that specifically wishes to catch failures in trial futures and retry
+the block should use C<try_repeat> specifically.
 
 =cut
 
-*try_repeat = \&repeat;
+sub try_repeat(&@)
+{
+   # defeat prototype
+   &repeat( @_, try => 1 );
+}
 
 =head2 $future = try_repeat_until_success { CODE } ...
 
@@ -373,7 +403,8 @@ sub try_repeat_until_success(&@)
    defined($args{while}) or defined($args{until})
       and croak "Cannot pass 'while' or 'until' to try_repeat_until_success";
 
-   try_repeat( $code, while => sub { shift->failure }, %args );
+   # defeat prototype
+   &try_repeat( $code, while => sub { shift->failure }, %args );
 }
 
 # Legacy name
@@ -405,7 +436,9 @@ The following named arguments are common to each C<fmap*> function:
 
 Provides the list of items to iterate over, as an C<ARRAY> reference.
 
-The referenced array may be modified by this operation.
+The referenced array will be modified by this operation, C<shift>ing one item
+from it each time. The can C<push> more items to this array as it runs, and
+they will be included in the iteration.
 
 =item generate => CODE
 
@@ -415,6 +448,9 @@ empty list to indicate it has no more items.
 
  ( $item ) = $generate->()
 
+This function will be invoked each time any previous item future has completed
+and may be called again even after it has returned empty.
+
 =item concurrent => INT
 
 Gives the number of item futures to keep outstanding. By default this value
@@ -447,9 +483,12 @@ below.
 # It can tell which circumstance by whether the slot itself is defined or not
 sub _fmap_slot
 {
-   my @args = my ( $slots, $idx, $code, $generator, $collect, $results, $return ) = @_;
+   my ( $slots, undef, $code, $generator, $collect, $results, $return ) = @_;
+
+   SLOT: while(1) {
+      # Capture args each call because we mutate them
+      my ( undef, $idx ) = my @args = @_;
 
-   while(1) {
       unless( $slots->[$idx] ) {
          # No item Future yet (case a), so create one
          my $item;
@@ -485,6 +524,17 @@ sub _fmap_slot
          $args[-1] = ( $return ||= $f->new );
          $f->on_done( sub { _fmap_slot( @args ) } );
          $f->on_fail( $return );
+
+         # Try looking for more that might be ready
+         my $i = $idx + 1;
+         while( $i != $idx ) {
+            $i++;
+            $i %= @$slots;
+            next if defined $slots->[$i];
+
+            $_[1] = $i;
+            redo SLOT;
+         }
          return $return;
       }
 
@@ -507,7 +557,7 @@ sub _fmap
    my %args = @_;
 
    my $concurrent = $args{concurrent} || 1;
-   my @slots = ( undef ) x $concurrent;
+   my @slots;
 
    my $results = [];
    my $future = $args{return};
@@ -524,7 +574,7 @@ sub _fmap
    }
 
    # If any of these immediately fail, don't bother continuing
-   foreach my $idx ( 0 .. $#slots ) {
+   foreach my $idx ( 0 .. $concurrent-1 ) {
       $future = _fmap_slot( \@slots, $idx, $code, $generator, $args{collect}, $results, $future );
       last if $future->is_ready;
    }
